{   obcc, a small oberoncompiler for the C16x-family
      based on Oberon-0 by N. Wirth

    Copyright 2011 Guido B. (user "guido-b" at http://www.mikrocontroller.net)

    obcg.pas, the generator, is part of obcc.
    
    obcc is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    obcc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with obcc.  If not, see <http://www.gnu.org/licenses/>.                }


unit obcg; { Der Codegenerator }

interface
uses obcs; { Scanner einbinden }

CONST //maxCode = 10000; maxRel = 200; NofCom = 16;
    (* class / mode *) g_Head = 0;
      g_Var = 1; g_Par = 2; g_Const = 3; g_Fld = 4; g_Typ = 5; g_Proc = 6; 
      g_ProcR = 7; g_SProc = 8;
      g_Int = 9; g_Reg = 10; g_Cond = 11; g_SFR = 12; g_String = 13;
      g_mod = 14;
    (* form *) g_Boolean = 0; g_Integer = 1; g_Array = 2; g_Record = 3; g_SFRF = 4; g_stringf = 5;

	maxvarcnt = 16; { in bytes }
	


TYPE g_Object = ^ObjDesc;
    g_Type = ^TypeDesc;
    Pequ = ^Tequ;

{ Die Strings sind in Sprocs }


    g_Item = RECORD
      mode, lev: INTEGER;
      i_type: g_Type;
      a, b, c, r: LONGINT;
      tok: String;
    END ;

    ObjDesc= RECORD
      class, lev: INTEGER;
      next, dsc: g_Object;
      o_type: g_Type;
      name: Ident;
      val: LONGINT;
      IsPara, exported: BOOLEAN;
    END ;

    TypeDesc = RECORD
      form, ref: INTEGER;
      fields: g_Object;
      base: g_Type;
      size, len: INTEGER
    END ;
    
    Tequ = RECORD
      name, origin:STRING;
      next: Pequ;
    END;
    
    Vekt = RECORD
      name: STRING;
      adr: LONGINT;
      used: BOOLEAN;
    END;
    
    

VAR boolType, intType, sfrType, stringType: g_Type;
    Lequ: Pequ;
    vektoren: ARRAY[0..71] OF Vekt;
    curlev, pc: INTEGER;
    //relx, cno: INTEGER;
    //entry, fixlist: LONGINT;
    //regs: SET; (* used registers *)
    //code: ARRAY [1..maxCode] OF LONGINT;
    //rel: ARRAY [1..maxRel] OF INTEGER;
    //comname: ARRAY [1..NofCom] OF Ident;  (*commands*)
    //comadr: ARRAY [1..NofCom] OF LONGINT;
    //mnemo: ARRAY [1..54,1..5] OF CHAR;  (*for decoder*)
    Nextreg, Lastreg, Lastop, ParCount, maxreg: INTEGER;
    nextlab: Longint;
    watchdog: BOOLEAN;
    ramstart, ramstop, ramtop, progstart, stackp, stacksize, contextp : LONGINT;
    waitstates: INTEGER;
    oplevel: INTEGER;
    isproc: BOOLEAN; { Procedure-Declaration }
    smallproc, minreg: BOOLEAN;
    forloop : BOOLEAN;
    vnr: INTEGER; vstr: STRING;
    IsInt: BOOLEAN; { procedure is an interrupt }
    Regs: Array[1..15] OF Longint;
    RegUsed : ARRAY[1..15] OF INTEGER;
    ConfFile: String;
    DPP: ARRAY[0..3] OF INTEGER;
    rammapped, impmod, write_used: BOOLEAN;
    ModName, MainName: Ident;
    ramoffset: INTEGER;
    DebugLevel, AscPort: Integer;

PROCEDURE RetReg(VAR x: g_Item);
PROCEDURE IncLevel(n: INTEGER);
PROCEDURE MakeConstItem(VAR x: g_Item; typ: g_Type; val: LONGINT);
PROCEDURE MakeStringItem(VAR x: g_Item; typ: g_Type; s: String);
PROCEDURE MakeItem(VAR x: g_Item; y: g_Object);
PROCEDURE Field(VAR x: g_Item; y: g_Object);   (* x := x.y *)
PROCEDURE Index(VAR x, y: g_Item);   (* x := x[y] *)
PROCEDURE Op1(op: INTEGER; VAR x: g_Item);   (* x := op x *)
PROCEDURE Op2(op: INTEGER; VAR x, y: g_Item);   (* x := x op y *)
PROCEDURE TestTrue(VAR x:g_Item);
PROCEDURE Relation(op: INTEGER; VAR x, y: g_Item);   (* x := x ? y *)
PROCEDURE Store(VAR x, y: g_Item); (* x := y *)
PROCEDURE StoreLv(VAR x: g_Item); {write back controllvariable after FOR}
PROCEDURE g_Parameter(VAR x: g_Item; ftyp: g_Type; class: INTEGER; sc: BOOLEAN);
PROCEDURE CJump(lab: LONGINT);
PROCEDURE BJump(lab: LONGINT);
PROCEDURE Call(VAR x: g_Object);
PROCEDURE IOCall(VAR x, y: g_Item);
PROCEDURE OpenWrite;
PROCEDURE PutWrite(x: g_Item; form: Integer);
PROCEDURE CloseWrite(nl: Boolean);
PROCEDURE Return(s: string; int: BOOLEAN);
PROCEDURE Open;
FUNCTION GetLabel:Longint;
PROCEDURE SetLabel(nlabel: Longint; s:string);
PROCEDURE NewLabel(nlabel: Longint; s:string);
PROCEDURE SetProcLabel(s: STRING; int: BOOLEAN);
PROCEDURE g_Inc(VAR x: g_Item);
PROCEDURE putverb;
PROCEDURE g_OpenProc(blksize: INTEGER; small: BOOLEAN);
PROCEDURE CallMain;
PROCEDURE Open_Gen;
PROCEDURE Init_Sproc(vsize:LONGINT);
PROCEDURE WriteVectors;
PROCEDURE End_MOD;

PROCEDURE load(VAR x: g_Item);
PROCEDURE InclMod(n: Ident);

implementation

USES Sprocs;

	  
	PROCEDURE GetReg(VAR r:integer);
	BEGIN {r := Nextreg; Nextreg := Nextreg + 1;
	  IF nextreg > maxreg THEN maxreg := nextreg;
	  if Nextreg = Lastreg then Mark('too many registers used!');}
	  r := nextreg;
	  IF oplevel > 1 THEN
	    WHILE ((RegUsed[r] > 0) OR (Regs[r] <> -1)) AND (r <= lastreg) DO
	      r := r + 1;
	  IF r > lastreg THEN r := nextreg;
	  WHILE (r <= lastreg) AND (RegUsed[r] > 0) DO
	    r:= r +1;
	  IF r > lastreg THEN Mark('too many registers used!')
	  ELSE RegUsed[r] := RegUsed[r] + 1;
	END;
	
	PROCEDURE RetReg(VAR x: g_Item);
	BEGIN IF ((x.r <= lastreg) AND (x.r > FSP)) OR (IsInt AND (x.r <= lastreg)) THEN BEGIN
	      {  if x.r = nextreg - 1 then Nextreg := nextreg - 1;}
	        RegUsed[x.r] := RegUsed[x.r] - 1;
	        IF x.mode = g_Reg THEN BEGIN
	          x.mode := g_VAR; x.r := g_PC; END
	        ELSE IF x.mode = g_Par THEN BEGIN 
	          IF smallproc THEN
	            x.r := 16 - (x.a DIV 2)
	          ELSE
	            x.r := FSP;
	        END
	      END;
	END;
	
	PROCEDURE g_Inc(VAR x: g_Item);
	  VAR reg, reg2: INTEGER; r: LONGINT;
	BEGIN
	  IF x.mode = g_Reg THEN BEGIN
	      writeln(Taus, chr(9), 'add', chr(9), 'R', x.r, ', #1');
	     IF (Not(smallproc)) AND (NOT(IsInt)) AND (oplevel = 0) THEN 
	     { Spezialbehandlung fr arrays ohne Optimierung }
	       writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', x.a, '], R', x.r);
	  END
	  ELSE IF (x.mode = g_Var) OR (x.mode = g_Par) THEN BEGIN
	    GetReg(reg); 
	    IF x.lev = 0 THEN BEGIN
	      r := ramstart + x.a - 2; 
	      writeln(Taus,char(9), 'mov', chr(9), 'R', reg, ', #1');
	      writeln(Taus, chr(9), 'add', chr(9), hexword(r), ', R', reg);
	    END
	    ELSE IF x.lev = curlev THEN BEGIN
	      IF smallproc THEN Begin { g_Par! }
	        writeln(Taus,char(9), 'mov', chr(9), 'R', reg, ', [R', x.r, ']');
	        writeln(Taus,char(9), 'add', chr(9), 'R', reg, ', #1');
	        writeln(Taus,char(9), 'mov', chr(9), '[R', x.r, '], R', reg);
	      END
	      ELSE BEGIN
	        writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', [R', FSP, '+#', x.a, ']');
	        IF x.mode = g_Par THEN BEGIN
	          getreg(reg2);
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg2, ', R', reg);
	          writeln(Taus,char(9), 'mov', chr(9), 'R', reg, ', [R', reg, ']');
	        END;
	        writeln(Taus,char(9), 'add', chr(9), 'R', reg, ', #1');
	        IF x.mode = g_Par THEN BEGIN
	          writeln(Taus,char(9), 'mov', chr(9), '[R', reg2, '], R', reg);
	          Regs[reg2] := -1; RegUsed[reg2] := 0;
	        END
	        ELSE
	          writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', x.a, '], R', reg);
	      END;
	      RegUsed[reg] := 0;
	      IF x.mode = g_Var THEN Regs[reg] := 65000 * x.lev + x.a;
	    END
	    ELSE Mark('level?');
	  END;
	END;
	
	PROCEDURE testrange(i:LONGINT);
	BEGIN IF (i > $FFFF) OR (i < 0) then 
	         Mark('value out of range');
	END;

	PROCEDURE load(VAR x: g_Item);
	  VAR r: LONGINT; reg: INTEGER;
	BEGIN (* erstmal x.mode # Reg: *)
	  IF x.mode = g_Const THEN BEGIN
             TestRange(x.a); GetReg(reg);
             writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', #', x.a); 
             x.mode := g_Reg; x.r := reg; Regs[reg] := -1;
          END
          ELSE IF x.mode = g_PAR THEN BEGIN	        
	     IF (smallproc) THEN  BEGIN
	      IF x.r >= lastreg THEN BEGIN
		Getreg(reg);
		writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', [R', x.r, ']');
		x.r := reg; (*x.mode := g_Reg;*) Regs[reg] := (65000 * x.lev) + x.a; END;{ Achtung: ist g_Reg o.k.? Ich glaube nicht!}
	     END
	     ELSE BEGIN 
	        Getreg(reg);
	        writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', [R', FSP, '+#', x.a, ']');
	        writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', [R', reg, ']');
	        x.r := reg; Regs[reg] := (65000 * x.lev) + x.a;
	     END
	  END
	  ELSE IF x.mode <> g_REG THEN BEGIN
	    x.r := g_PC;
	    IF oplevel > 0 Then
	    { Nicht fr Level-0-Variablen, sonst gibt es ein Not-Volatile-Problem!.}
	      IF x.lev <> 0 THEN BEGIN
	        reg := nextreg;
	        WHILE (reg <= 15) AND (Regs[reg] <> ((65000 * x.lev) + x.a)) DO 
	          reg := reg + 1;
	        IF (reg < 16) AND (Regs[reg] = ((65000 * x.lev) + x.a)) THEN BEGIN
	          x.r := reg; X.mode := g_Reg; {nextreg := reg + 1;}
	          RegUsed[x.r] := RegUsed[x.r] + 1; END 
	      END;	    
	    IF (x.r = g_PC)  THEN BEGIN
	      IF (x.mode = g_Var) OR (x.mode = g_Par) THEN BEGIN
	        IF x.lev = 0 THEN BEGIN
	          r := ramstart + x.a - 2; 
	          GetReg(reg); 
	          writeln(Taus,char(9), 'mov', chr(9), 'R', reg, ', (', hexword(r), ' + roffs)');
	          x.r := reg; x.mode := g_Reg; Regs[reg] := (65000 * x.lev) + x.a;
	        END
	        ELSE IF x.lev = curlev THEN BEGIN
	          IF x.mode = g_Var THEN BEGIN
	            IF NOT smallproc THEN BEGIN
	              Getreg(reg); //r := x.a - 2;
	              writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', [R', FSP, '+#', x.a, ']');
	              x.r := reg; x.mode := g_Reg; Regs[reg] := (65000 * x.lev) + x.a;
	            END;
	          END	          
	        END
	        ELSE Mark('level!');
	      END
            END;
          END;
	END {load};
	
	FUNCTION GetLabel:Longint;
	BEGIN GetLabel := NextLab;
	  NextLab := NextLab + 1;
	END;

	PROCEDURE SetLabel(nlabel: Longint; s:string); { Nach z.B. jmp ohne Doppelpunkt }
	  VAR i, j: INTEGER;
	BEGIN i := 10000; J := 5;
	  Write(Taus,modname);
	  WHILE (i > 1) AND (nlabel < i) DO BEGIN
	    write(Taus,'0'); j:=j-1; i := i div 10; END;
	  write(Taus, nlabel:j);
	  if s <>'' then writeln(Taus, chr(9), chr(9), chr(9), '; ', s)
	  else writeln(Taus);
	END;
	
	PROCEDURE NewLabel(nlabel: Longint; s:string); { als Sprungziel mit Doppelpunkt }
	  VAR i, j: INTEGER;
	BEGIN i := 10000; J := 5;
	  Write(Taus, modname);
	  WHILE (i > 1) AND (nlabel < i) DO BEGIN
	    write(Taus,'0'); j:=j-1; i := i div 10; END;
	  write(Taus, nlabel:j);
	  if s <>'' then writeln(Taus, ':', chr(9), chr(9), chr(9), chr(9), chr(9), '; ', s)
	  else writeln(Taus, ':');
	END;
	
	PROCEDURE CheckStack;
	Var lab: LongInt;
	Begin
	  { Stack checken, mindestens 44 Bytes mssen fr
	    einen Interrupt brigbleiben (besser 82?) + 2 Bytes Reserve. }
	    lab := GetLabel;
	    writeln(Taus, '; CheckStack');
	    writeln(Taus, chr(9), 'mov', chr(9), 'R1, SP');
	    writeln(Taus, chr(9), 'mov', chr(9), 'R2, CP');
	    writeln(taus, chr(9), 'sub', chr(9), 'R1, R2');
	    Writeln(Taus, chr(9), 'cmp', chr(9), 'R1, #46');
	    write(Taus, chr(9), 'jmp', chr(9), 'UGE, ');
	    setlabel(lab, '');
	    writeln(Taus, chr(9), 'bset', chr(9), 'STKOF');
	    NewLabel(lab, 'CheckStack');
	End; {CheckStack}
	
	PROCEDURE CheckIndex(max, regind:Integer);
	VAR lab: LongInt;
	Begin { Arryindex checken }
	  lab := GetLabel;
	  writeln(Taus, '; CheckIndex');
	  writeln(taus, chr(9), 'cmp', chr(9), 'R', regind, ', #0');
	  write(taus, chr(9), 'jmp', chr(9), 'NE, '); setlabel(lab,'');
	  writeln(taus, chr(9), 'bset', chr(9), 'ILLOPA');
	  NewLabel(lab, ''); lab := GetLabel;
	  writeln(taus, chr(9), 'cmp', chr(9), 'R', regind, ', #', max);
	  write(taus, chr(9), 'jmp', chr(9), 'ULE, '); setlabel(lab,'');
	  writeln(taus, chr(9), 'bset', chr(9), 'ILLOPA');
	  NewLabel(lab, 'CheckIndex');
	End; {CheckIndex}
	
	PROCEDURE CheckFlag(bed:String);
	VAR lab: LongInt;
	Begin lab := GetLabel;
	  writeln(Taus, '; CheckFlag');
	  write(taus, chr(9), 'jmp', chr(9), bed, ', '); SetLabel(lab, '');
	  writeln(Taus, chr(9), 'bset', chr(9), 'ILLINA');
	  NewLabel(lab, 'CheckFlag');
	End; {CheckFlag}

	PROCEDURE SetProcLabel(s: STRING; int: BOOLEAN);
	BEGIN IF NOT int THEN s := modname + s;
	  writeln(Taus, s, ':');
	  IF int THEN BEGIN
	    writeln(Taus, chr(9), 'add', chr(9), 'CP, #32');
	    writeln(Taus, chr(9), 'nop');
	    If DebugLevel > 1 Then CheckStack;
	  END;
	END;
	
	FUNCTION IsPot(VAR k:Longint; a:LONGINT):BOOLEAN;
	  VAR j: INTEGER;
	BEGIN j := a; k := 0;
	  WHILE NOT ODD(j) DO BEGIN
	    j := j DIV 2; k := k + 1;
	  END;
	  IF j = 1 THEN IsPot := TRUE
	  ELSE IsPot := FALSE;
	END;


	PROCEDURE IncLevel(n: INTEGER);
	  BEGIN curlev := curlev + n;
	  END {IncLevel};

	PROCEDURE MakeConstItem(VAR x: g_Item; typ: g_Type; val: LONGINT);
	  BEGIN x.mode := g_Const; x.i_type := typ; x.a := val;  x.lev := curlev; x.r := g_PC;
	  END {MakeConstItem};
	  
	PROCEDURE MakeStringItem(VAR x: g_Item; typ: g_Type; s: String);
	  VAR nstring: stringp;
	BEGIN x.a := nextstring; nextstring := nextstring + 1;
	  NEW(nstring); nstring^.nr := x.a; nstring^.inh := s;
	  nstring^.next := stringl; stringl := nstring;
	  x.mode := g_String; x.i_type := typ; x.lev := curlev;
	END;
	

	PROCEDURE MakeItem(VAR x: g_Item; y: g_Object);
	  VAR Fequ:Pequ;
	BEGIN x.mode := y^.class; x.i_type := y^.o_type; x.lev := y^.lev; x.a := y^.val;
	  IF y^.lev = 0 THEN BEGIN
	    x.r := g_PC; 
	  END
	  ELSE IF y^.lev = curlev THEN BEGIN
	    IF x.mode = g_Var THEN BEGIN
	      IF smallproc THEN BEGIN
	        x.mode := g_Reg; x.r := (16 - x.a DIV 2); END
	      ELSE BEGIN
	        x.r := FSP; x.a := x.a - 2; END;
	    END
	    ELSE IF x.mode = g_Par THEN BEGIN
	      IF smallproc THEN 
	        x.r := 16 - (x.a DIV 2)
	      ELSE BEGIN
	        x.r := FSP; x.a := x.a - 2;
	      END;
	    END;
	  END
	  ELSE IF (y^.name <> 'TRUE') AND (y^.name <> 'FALSE') 
	    AND (y^.class <> g_Proc) AND (y^.class <> g_ProcR) AND (y^.class <> g_Sproc) THEN 
	    BEGIN 
	      Mark('level?!'); x.r := 0; 
	  END;
	  IF y^.class = g_SFR THEN BEGIN
	    x.tok := y^.name; Fequ := Lequ;{ Alias-Name? }
	    WHILE Fequ <> NIL DO (* TODO: Levelkontrolle dazu? *)
	      IF Fequ^.name = x.tok THEN BEGIN
	        x.tok := Fequ^.origin; Fequ := NIL; END
	      ELSE Fequ := Fequ^.next;
	  END;    
	END {MakeItem};

	PROCEDURE Field(VAR x: g_Item; y: g_Object);   (* x := x.y *)
	VAR offs, reg: INTEGER;
	  BEGIN 
	    IF x.i_type^.form = g_Record THEN BEGIN
	      offs := 2 - x.i_type^.size + y^.val; 
	      IF x.mode = g_PAR THEN BEGIN
	        GetReg(reg); 
	        IF smallproc THEN BEGIN
 	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', R', x.r);
	        END
	        ELSE BEGIN
		  writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', [R', FSP, '+#', x.a, ']');
	        END;
	        writeln(Taus, chr(9), 'add', chr(9), 'R', reg, ', #', y^.val + 2);
	        x.r := reg;
	      
	      END
	      ELSE
	        x.a := x.a + offs;
	      END
	    ELSE Mark('no record!');
	  END {Field};

	PROCEDURE Index(VAR x, y: g_Item);   (* x := x[y] *)
	  VAR reg1, reg2, reg3: INTEGER; offs,l: LONGINT;
	  BEGIN 
	    IF x.lev = 0 THEN BEGIN
	      getreg(reg1); x.r := reg1; x.mode := g_Reg;
	      IF y.mode = g_Const THEN BEGIN 
	        IF (y.a <= 0) OR (y.a > x.i_type^.len) THEN  Mark('invalid index');
	        offs := ramstart + x.a - x.i_type^.size + (y.a - 1) * x.i_type^.base^.size	        
	      END
	      ELSE BEGIN 
	        offs := ramstart + (x.a - 2) - x.i_type^.size;
	      END;
	      writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', #(', hexword(offs), ' + roffs)');
	    END
	    ELSE IF smallproc THEN BEGIN
	      getreg(reg1); 
	      writeln(Taus, chr(9), 'mov', chr(9), 'R', reg1, ', R', x.r);
	      x.r := reg1;
	      IF y.mode = g_Const THEN Begin
	        writeln(Taus, chr(9), 'add', chr(9), 'R', reg1, ', #', y.a  * x.i_type^.base^.size);
	        IF (y.a <= 0) OR (y.a > x.i_type^.len) THEN  Mark('invalid index');
	      End;
	    END
	    ELSE BEGIN { lange Prozedur }	
	      IF x.mode = g_Par THEN BEGIN
		GetReg(reg1);
		Writeln(Taus, chr(9), 'mov', chr(9), 'R', reg1, ', [R', FSP, '+#', x.a, ']');
		x.r := reg1; x.mode := g_Reg;
		IF y.mode = g_Const THEN Begin
		  IF (y.a <= 0) OR (y.a > x.i_type^.len) THEN  Mark('invalid index');
	          writeln(Taus, chr(9), 'add', chr(9), 'R', reg1, ', #', y.a  * x.i_type^.base^.size);
	        END;
	      END
	      ELSE
	      IF y.mode = g_Const THEN BEGIN
	          IF (y.a <= 0) OR (y.a > x.i_type^.len) THEN  Mark('invalid index');
	          x.r := FSP;
	          offs := x.a - x.i_type^.size + 2 + (y.a - 1) * x.i_type^.base^.size; 
	          x.a := offs; END
	      ELSE BEGIN
	          offs := x.a - x.i_type^.size + 2;
	          getreg(reg1); x.r := reg1; x.mode := g_Reg;
	        //writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', [R', FSP, ']');
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', R', FSP);  
	          writeln(Taus, chr(9), 'add', chr(9), 'R', x.r, ', #', offs);
	          x.a := 0;
	      END;
	    END;	    
	    IF (y.mode = g_Var) THEN BEGIN
	        load(y); 
	        IF RegUsed[y.r] > 1 THEN BEGIN
	          getreg(reg2); 
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg2, ', R', y.r);
	          RetReg(y); y.r := reg2;
	        END;
	    END
	    ELSE IF y.mode = g_Reg THEN BEGIN 
	      GetReg(reg2);
	      writeln(Taus, chr(9), 'mov', chr(9), 'R', reg2, ', R', y.r);
	      y.r := reg2;
	    END;
	    IF y.mode <> g_Const THEN BEGIN
	    // Hier Check auf legalen Index einfgen! TODO
	      If DebugLevel > 1 Then CheckIndex(x.i_type^.len, y.r);
 	      offs := x.i_type^.base^.size;
	      IF IsPot(l,offs) THEN
	        writeln(Taus, chr(9), 'shl', chr(9), 'R', y.r, ', #', l)
	      ELSE BEGIN
	        GetReg(reg3);
	        writeln(Taus, chr(9), 'mov', chr(9), 'R', reg3, ', #', offs);
	        writeln(Taus, chr(9), 'mulu', chr(9), 'R', y.r, ', R', reg3);
	        writeln(Taus, chr(9), 'mov', chr(9), 'R', y.r, ', MDL');
	        y.r := reg3; RetReg(y); y.r := reg2;
	      END;
	      writeln(Taus, chr(9), 'add', chr(9), 'R', x.r, ', R', y.r);
	      reg3 := y.r; y.r := reg2; retreg(y); y.r := reg3;
	    END;
	  END {Index};
	  
	PROCEDURE CJump(lab:longint);
	  var op1:string;
	  BEGIN
	    CASE Lastop of
	      eql_: op1 := 'NE';
	      neq_: op1 := 'EQ';
	      lss_: op1 := 'UGE';
	      geq_: op1 := 'ULT';
	      leq_: op1 := 'UGT';
	      gtr_: op1 := 'ULE';
	    END;
	    write(Taus, chr(9), 'jmp', chr(9), op1, ', ');
	    SetLabel(lab, '');
	  END {CJump};
  
	PROCEDURE Op1(op: INTEGER; VAR x: g_Item);   (* x := op x *)
	  VAR lab1, lab2: Longint;
	  BEGIN 
	    CASE op OF
	      minus_: Mark(' no negative integers!');
	      not_: BEGIN if x.mode <> g_Reg then load(x);
	              writeln(Taus, chr(9), 'xor', chr(9), 'R', x.r, ', #1'); 
	            END;
	      or_: BEGIN 
	             lab1 := Getlabel; x.a := Getlabel;
	             CJump(lab1);
	             write(Taus, chr(9), 'jmp', chr(9)); Setlabel(x.a, 'OR passed'); 
	             NeWlabel(lab1, ''); x.mode := g_Cond;
	           END;
	      and_: BEGIN 
	             lab1 := Getlabel; lab2 := Getlabel; x.a := Getlabel;
	             CJump(lab1);
	             write(Taus, chr(9), 'jmp', chr(9)); Setlabel(lab2, 'AND passed');
	             NeWlabel(lab1, ''); 
	             write(Taus, chr(9), 'jmp', chr(9)); Setlabel(x.a, 'AND failed');
	             Newlabel(lab2, ' AND next');
	             x.mode := g_Cond;
	            END;
	      ELSE Mark('operator?');
	    END;
	  END {Op1};
	  
	

	PROCEDURE Op2(op: INTEGER; VAR x, y: g_Item);   (* x := x op y *)
	  VAR lab1, lab2, n: LONGINT; reg: INTEGER;
	  BEGIN 
	    IF x.mode = g_Cond THEN BEGIN
	        IF op = or_ THEN BEGIN
	          lab1 := GetLabel; lab2 := GetLabel; CJump(lab1); //CJump(x.a);
	          Getreg(reg);
	          Newlabel(x.a, ' OR passed');
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', #1');
	          write(Taus, chr(9), 'jmp', chr(9)); setlabel(lab2, '');
	          NewLabel(lab1, ' OR failed');
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', #0');
	          Newlabel(lab2, ' OR done'); 
	          lastop := neq_; x.r := reg; retreg(x);
	        END
	        ELSE IF op = and_ THEN BEGIN
	          lab1 := GetLabel; 
	          CJump(x.a); Getreg(reg);
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', #0');
	          write(Taus, chr(9), 'jmp', chr(9)); setlabel(lab1, '');
	          NewLabel(x.a, 'AND failed');
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', #1');
	          Newlabel(lab1, ' AND done'); 
	          lastop := eql_; x.r := reg; retreg(x); 
	        END
	        ELSE Mark('operator?');
	    END { conditional}
	    ELSE IF (x.i_type^.form = g_Integer) AND (y.i_type^.form = g_Integer) THEN BEGIN
	      IF (x.mode = g_Const) AND (y.mode = g_Const) THEN
	        CASE op OF
	          plus_:  BEGIN x.a := x.a + y.a;
	          	  if x.a > $FFFF THEN Warn('overflow'); END;
	          minus_: BEGIN x.a := x.a - y.a;
	          	  if x.a < 0 THEN Warn('underflow'); END;
	          times_: BEGIN x.a := (x.a * y.a); // MOD $10000;
			if x.a > $FFFF then Warn('overflow'); END;
	          div_:   x.a := x.a DIV y.a;
	          mod_:   x.a := x.a MOD y.a;
	          ELSE mark('bad type!');
	        END
	      ELSE IF (y.mode = g_Const) AND IsPot(n, y.a) AND (op IN [times_, div_, mod_]) THEN BEGIN
	        IF x.mode <> g_Reg THEN load(x);
	        {ELSE IF (x.r > lastreg) THEN BEGIN
	          GetReg(reg); 
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', R', x.r);
	          x.r := reg;
	        END;}
	        CASE op OF
	          times_: 
	            IF y.a = 0 THEN
	              writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', #0')
	            ELSE BEGIN
	              writeln(Taus, chr(9), 'shl', chr(9), 'R', x.r, ', #', n);
	              If DebugLevel > 1 Then CheckFlag('NC');
	            End;
	          div_:
	            IF y.a = 0 THEN
	            Mark('div. by zero!')
	              //writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', #$FFFF')
	            ELSE If n > 0 Then
	              writeln(Taus, chr(9), 'shr', chr(9), 'R', x.r, ', #', n);
	          mod_:
	            IF y.a = 0 THEN
		      Mark('div. by zero!')
	              //writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', #$FFFF')
	            ELSE 
	              writeln(Taus, chr(9), 'and', chr(9),'R', x.r, ', #', hexword(y.a - 1));
	        END;
	      END
	      ELSE BEGIN
	        IF x.mode <> g_Reg THEN load(x);
	        //IF y.mode <> g_Reg THEN load(y);
	        {ELSE IF (x.r > lastreg) THEN BEGIN
	          GetReg(reg);
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', R', x.r);
	          x.r := reg;
	        END;}
	        IF (y.mode <> g_Const) OR (op = times_) then load(y);
	        CASE op OF
	          plus_:  write(Taus, chr(9), 'add', chr (9), 'R', x.r, ', ');
	          minus_: write(Taus, chr(9), 'sub', chr (9), 'R', x.r, ', ');
	          andb_ :  write(Taus, chr(9), 'and', chr (9), 'R', x.r, ', ');
	          orb_:    write(Taus, chr(9), 'or', chr (9), 'R', x.r, ', ');
	          xorb_:   write(Taus, chr(9), 'xor', chr (9), 'R', x.r, ', ');
	          times_: write(Taus, chr(9), 'mulu', chr (9), 'R', x.r, ', ');
	          div_, mod_: IF (y.mode <> g_Reg) then load(y);
	          ELSE mark('bad type!');
	        END;
	        CASE op OF
	          plus_, minus_: BEGIN
	              IF (y.mode <> g_Reg) THEN writeln(Taus, '#', y.a)
	          		 ELSE writeln(Taus, 'R', y.r);
		      IF DebugLevel > 1 Then CheckFlag('NC');
		    End;
	          andb_, orb_,xorb_: 
		      IF (y.mode <> g_Reg) THEN writeln(Taus, '#', y.a)
	          		 ELSE writeln(Taus, 'R', y.r);
	          times_: BEGIN writeln(Taus, 'R', y.r);
				IF DebugLevel > 1 Then CheckFlag('NV');
				IF DPP[3] <> 3 THEN 
				  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
	          		writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', MDL'); END;
	          div_: BEGIN 
			      IF DPP[3] <> 3 THEN 
				  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
			      writeln(Taus, chr(9), 'mov', chr(9), 'MDL, R', x.r);
	          	      writeln(Taus, chr(9), 'divu', chr(9), 'R', y.r);
	          	      IF DebugLevel > 1 Then CheckFlag('NV');
	          	      IF DPP[3] <> 3 THEN 
				  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
	          	      writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', MDL'); END;
	          mod_: BEGIN 
			      IF DPP[3] <> 3 THEN 
				  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
			      writeln(Taus, chr(9), 'mov', chr(9), 'MDL, R', x.r);
	          	      writeln(Taus, chr(9), 'divu', chr(9), 'R', y.r);
	          	      IF DebugLevel > 1 Then CheckFlag('NV');
	          	      IF DPP[3] <> 3 THEN 
				  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
	          	      writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', MDH'); END;
	          ELSE 
	        END;
	        IF (y.mode = g_Reg) THEN Retreg(y); 
	      END;
	    END { Integer }
	    ELSE IF (x.i_type^.form = g_Boolean) AND (y.i_type^.form = g_Boolean) THEN BEGIN
	        load(x);
	        IF (y.mode <> g_Const) then load(y);
	        CASE op OF
	          and_: write(Taus, chr(9), 'and', chr(9), 'R', x.r, ', ');
	          or_:  write(Taus, chr(9), 'or', chr(9), 'R', x.r, ', ');
	          ELSE mark('operator?');
	        END;
	        IF (y.mode = g_Const) THEN writeln(Taus, '#', y.a)
	        ELSE BEGIN 
	          writeln(Taus, 'R', y.r); Retreg(y); END
	    END { Boolean }
	  END {Op2};
	  
	PROCEDURE TestTrue(VAR x:g_Item);
	BEGIN 
	  IF x.mode <> g_Reg THEN load(x);
	  writeln(Taus, chr(9), 'cmp', chr(9), 'R', x.r, ', #', 1);
	  Retreg(x);
	  Lastop := eql_;
	END;

	PROCEDURE Relation(op: INTEGER; VAR x, y: g_Item);   (* x := x ? y *)
	  BEGIN Lastop := op; 
	    IF (x.i_type^.form <> g_Integer) OR (y.i_type^.form <> g_Integer) THEN Mark('bad type!') 
	    ELSE BEGIN  {g_Par ist noch falsch!}
	      IF (y.mode = g_Par) AND (NOT smallproc) THEN BEGIN
	        load(y);
	      END
	      ELSE IF (y.mode = g_Var) AND (y.lev <>0) THEN
	        load(y);
	      IF x.mode = g_Par THEN BEGIN
	        IF smallproc THEN BEGIN
	          IF x.r > lastreg THEN load(x);
	          write(Taus, chr(9), 'cmp', chr(9), 'R', x.r);
	        END
	        ELSE Begin
	          load(x);
	          write(Taus, chr(9), 'cmp', chr(9), 'R', x.r);
	        END;
	      END
	      ELSE BEGIN
	        IF x.mode <> g_Reg THEN load(x);
	        write(Taus, chr(9), 'cmp', chr(9), 'R', x.r);
	      END;
	      CASE y.mode OF
	        g_Const: writeln(Taus, ', #', y.a);
	        g_Reg: writeln(Taus, ', R', y.r);
	        g_Var: BEGIN
	          IF y.lev = 0 then BEGIN
	            writeln(Taus, ', (', hexword(ramstart + y.a - 2), ' + roffs)')
	          {ELSE IF y.lev = curlev THEN BEGIN
	            IF smallproc THEN 
	              writeln(Taus, ', R', y.r)
	            ELSE BEGIN
	              //load(y);
	              writeln(Taus, ', R', y.r);
	              retreg(y);
	            END;}
	          END ELSE Mark('level!');
	        END;
	        g_Par: BEGIN 
	          IF smallproc THEN { wenn nicht, ist y geladen! Also sollte dies fr beide Flle funktionieren? }
	            writeln(Taus, ', [R', y.r, ']')
	          ELSE BEGIN
	          { TODO }
	          END;
	        END;
	        ELSE Mark('mode?');
	      END; { case }
	      x.c := op;
	      IF NOT(forloop) THEN BEGIN
	        IF (x.mode = g_Reg) OR (x.mode = g_Par) THEN Retreg(x);
	        IF (y.mode = g_Reg) OR (y.mode = g_Par) THEN Retreg(y);
	      END;
	    END;
	  END {Relation};
  
	PROCEDURE Store(VAR x, y: g_Item); (* x := y *)
	  VAR r1: LONGINT; reg, reg2, zform, qform: INTEGER;
	  
	  PROCEDURE DecY(x,y: g_Item);
	    var i: INTEGER; r1, varval: LONGINT;
	  BEGIN
	      varval := (65000 * x.lev) + x.a;
	      IF x.mode = g_Reg THEN BEGIN 
		FOR i := nextreg TO lastreg do
	          IF Regs[i] = varval THEN Regs[i] := -1;
	        
	      END;
	      CASE y.mode OF
	          g_Reg: BEGIN
	            IF (y.i_type^.form = g_Array) AND (x.mode = g_Reg) THEN
	              writeln(Taus,', [R', y.r, ']')
	            ELSE
	              writeln(Taus,', R', y.r);
	            FOR i := nextreg TO lastreg do
		      IF Regs[i] = varval THEN Regs[i] := -1;
	            Regs[y.r] := varval;
	            Retreg(y); 
	          END;
	          g_Const: BEGIN
	            writeln(Taus, ', #', y.a); 
	            if y.mode = g_Reg THEN retreg(y);
	          END;
	          g_Var: BEGIN
	            IF (y.lev = 0) THEN BEGIN
	              r1 := ramstart + y.a - 2;
	              writeln(Taus, ', (',hexword(r1), ' + roffs)');
	            END
	            ELSE Mark('level!');
	          END;
	          g_Par: BEGIN 
	            IF smallproc THEN BEGIN
	              { ??? IF y.r < lastreg THEN
	                writeln(Taus, ', R', y.r)
	              ELSE} {wozu war das gut??? Strt bei Arrays als var. Parameter!}
	                writeln(Taus, ', [R', y.r, ']');
	               IF y.r < lastreg THEN retreg(y); 
	            END
	            ELSE BEGIN
	                writeln(Taus,', R', y.r);
	                RetReg(y);
	            END;
	          END;
	          ELSE mark('mode not Var?');
	      END;
	  END;
	  
	  BEGIN { store }
	    IF x.i_type^.form IN [g_Array, g_Record] THEN zform := x.i_type^.base^.form
	    ELSE zform := x.i_type^.form;
	    IF y.i_type^.form IN [g_Array, g_Record] THEN qform := y.i_type^.base^.form
	    ELSE qform := y.i_type^.form;
	    //IF (x.i_type^.form < 4) AND (y.i_type^.form < 4) AND (x.i_type^.base = y.i_type^.base) THEN BEGIN
	    IF (x.i_type^.form < 4) AND (y.i_type^.form < 4) AND (zform = qform) THEN BEGIN
	      IF (x.mode = g_Var) OR (x.mode = g_Reg) THEN BEGIN
	        IF x.lev = 0 then BEGIN
	          load(y);
	          IF x.mode = g_Var THEN BEGIN
	            r1 := ramstart + x.a - 2; { direct Variable }
	            write(Taus, chr(9), 'mov', chr(9), '(', hexword(r1), ' + roffs)'); END
	          ELSE BEGIN { Array }
	            write(Taus, chr(9), 'mov', chr(9), '[R', x.r, ']');
	          END;
	        END
	        ELSE IF x.lev = curlev THEN BEGIN
	          IF smallproc THEN BEGIN
	            if x.r = y.r THEN exit;
	            write(Taus, chr(9), 'mov', chr(9), 'R', x.r);
	          END
	          ELSE BEGIN
	            load(y); 
	            IF y.i_type^.form = g_Array THEN
	              writeln(Taus, chr(9), 'mov', chr(9), 'R', y.r, ', [R', y.r, ']');
	            IF x.i_type^.form = g_Array THEN	            
	              write(Taus, chr(9), 'mov', chr(9), '[R', x.r, '+#', x.a, ']')
	            ELSE	              
	              write(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', x.a, ']');	            
	          END;
	        END
	        ELSE BEGIN Mark('level!'); exit; END;
	        DecY(x,y); if x.mode = g_Reg THEN retreg(x);
	      END  
	      ELSE IF x.mode = g_Par THEN BEGIN
	        IF smallproc THEN BEGIN
	          IF y.mode = g_Const THEN BEGIN
	            getreg(reg);
	            writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', #', y.a);
	            writeln(Taus, chr(9), 'mov', chr(9), '[R', x.r, '], R',reg);
	            y.r := reg; retreg(y);
	          END
	          ELSE BEGIN
	            write(Taus, chr(9), 'mov', chr(9), '[R', x.r, ']');
	            DecY(x, y);
	          END;
	        END
	        ELSE BEGIN
		  IF x.r = FSP THEN BEGIN { sonst hat das field bzw. index schon erledigt }
		    getreg(reg); x.r := reg;
		    writeln(Taus, chr(9), 'mov', char(9), 'R', reg, ', [R', FSP, '+#', x.a, ']');
		  END;
	          IF (y.r > lastreg) OR (y.r = FSP) THEN load(y);
	            writeln(Taus, chr(9), 'mov', chr(9), '[R',x.r, '], ', 'R',  y.r);
	            RetReg(y);  
	        END;
	        IF x.r < lastreg THEN RetReg(x);
	      END
	      ELSE Mark('illegal assignement');
            END 
            ELSE IF ((x.i_type^.form = g_SFRF) AND (y.i_type^.form = g_Integer)) THEN BEGIN
              IF y.mode = g_Const THEN BEGIN
                IF x.tok[1] = 'e' THEN BEGIN
                  delete(x.tok, 1, 1);
                  IF DPP[3] = 3 THEN
		    writeln(Taus, chr(9), 'extr', chr(9), '#1')
		  ELSE
		    writeln(Taus, chr(9), 'extpr', chr(9), '#3, #1'); END;
                writeln(Taus, chr(9), 'mov', chr(9), x.tok, ', #', y.a)
              END
              ELSE IF y.mode = g_Reg THEN BEGIN
                IF x.tok[1] = 'e' THEN BEGIN
                  delete(x.tok, 1, 1);
                  IF DPP[3] = 3 THEN
		    writeln(Taus, chr(9), 'extr', chr(9), '#1')
		  ELSE
		    writeln(Taus, chr(9), 'extpr', chr(9), '#3, #1')
		END
		ELSE IF DPP[3] <> 3 THEN
		  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');
                writeln(Taus, chr(9), 'mov', chr(9), x.tok, ', R', y.r);
                IF y.r <= lastreg THEN RetReg(Y);
              END              
              ELSE IF (y.mode = g_Par) AND (smallproc) THEN BEGIN
                IF x.tok[1] = 'e' THEN BEGIN
                  delete(x.tok, 1, 1);
                  IF DPP[3] = 3 THEN
		    writeln(Taus, chr(9), 'extr', chr(9), '#1')
		  ELSE
                    writeln(Taus, chr(9), 'extpr', chr(9), '#3, #1'); 
                END
                ELSE IF DPP[3] <> 3 THEN
		  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');
                writeln(Taus, chr(9), 'mov', chr(9), x.tok, ', [R', y.r, ']')
              END              
              ELSE BEGIN { g_Var }
                load(y);
                IF x.tok[1] = 'e' THEN BEGIN
                  delete(x.tok, 1, 1);
                  IF DPP[3] = 3 THEN
		    writeln(Taus, chr(9), 'extr', chr(9), '#1')
		  ELSE
                    writeln(Taus, chr(9), 'extpr', chr(9), '#3, #1'); 
                END
                ELSE IF DPP[3] <> 3 THEN
		  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');
                writeln(Taus, chr(9), 'mov', chr(9), x.tok, ', R', y.r); 
                retreg(y); END;
            END            
            ELSE IF ((x.i_type^.form = g_Integer) AND (y.i_type^.form = g_SFRF)) THEN BEGIN
              IF x.mode = g_Reg THEN BEGIN
                IF y.tok[1] = 'e' THEN BEGIN
                  delete(y.tok, 1, 1);
                  IF DPP[3] = 3 THEN
		    writeln(Taus, chr(9), 'extr', chr(9), '#1')
		  ELSE
                    writeln(Taus, chr(9), 'extpr', chr(9), '#3, #1'); 
                END
                ELSE IF DPP[3] <> 3 THEN
		  writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');
                writeln(Taus, chr(9), 'mov', chr(9), 'R', x.r, ', ', y.tok)
              END              
              ELSE IF x.mode = g_Var THEN BEGIN
                IF y.tok[1] = 'e' THEN BEGIN
                  delete(y.tok, 1, 1);
                  writeln(Taus, chr(9), 'extr', chr(9), '#1'); END;
                IF x.lev = 0 THEN BEGIN 
                  r1 := ramstart + x.a - 2;
                  writeln(Taus, chr(9), 'mov', chr(9), '(', hexword(r1), ' + roffs), ', y.tok);
                END
                ELSE IF (x.lev = curlev) THEN BEGIN
                  IF NOT smallproc THEN BEGIN
                    GetReg(reg);
                    IF y.tok[1] = 'e' THEN BEGIN
		      delete(y.tok, 1, 1);
		      IF DPP[3] = 3 THEN
			writeln(Taus, chr(9), 'extr', chr(9), '#1')
		      ELSE
			writeln(Taus, chr(9), 'extpr', chr(9), '#3, #1'); 
		    END
		    ELSE IF DPP[3] <> 3 THEN
		      writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');                                        
		    writeln(Taus,chr(9), 'mov', chr(9), 'R', reg, ', ', y.tok);	        
	            writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', x.a, '], R', reg);
	            y.r := reg; Retreg(y);
	          END;
                END
                ELSE Mark('level!');
              END
              ELSE IF x.mode = g_Par THEN BEGIN
                IF x.tok[1] = 'e' THEN BEGIN
                  delete(x.tok, 1, 1);
                  writeln(Taus, chr(9), 'extr', chr(9), '#1'); END;
                IF smallproc THEN 
	          writeln(Taus, chr(9), 'mov', chr(9), '[R', x.r, '], ', y.tok)
	        ELSE BEGIN
	          Getreg(reg);
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg, ', ', y.tok);
	          GetReg(reg2);
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', reg2, ', [R', FSP, '+#', x.a, ']');
	          writeln(Taus, chr(9), 'mov', chr(9), '[R', reg2, '], R', reg);
	          y.r := reg; RetReg(y); x.r := reg2; RetReg(x);
	        END;
              END
              ELSE Mark(' mode not Var?');
            END
      	    ELSE Mark('incompatible assignement');
	  END {Store};	
	  
	PROCEDURE StoreLv(VAR x: g_Item); {write back controllvariable after FOR}
	BEGIN 
	  IF oplevel > 0 THEN
	    IF x.lev = 0 THEN BEGIN
	      x.a := x.a - 2;  
	      writeln(Taus, chr(9), 'mov', chr(9), '(',hexword(ramstart + x.a), ' + roffs), R', x.r);
	      retreg(x);
	    END
	    ELSE 
	      IF (NOT smallproc) AND (NOT IsInt) THEN BEGIN
	        writeln(Taus, chr(9), 'mov', chr(9),'[R', FSP, '+#', x.a, '], R', x.r);
	        retreg(x);
	      END;
	END {StorLv};
	
	PROCEDURE putverb;
	  VAR c: CHAR; s: STRING; weiter, d : BOOLEAN;
	BEGIN writeln(Taus, '; verbatim');
	  weiter := TRUE;
	  REPEAT
	    s := ''; d := Readmod(c);
	    WHILE c <> chr(13) DO BEGIN
	     s := s + c; d := Readmod(c);
	    END;
	    IF pos('END_VERBATIM',s) <> 0 THEN weiter := FALSE
	    ELSE writeln(Taus,s);
	  UNTIL NOT weiter;
	  writeln(Taus, '; end_verbatim');
	END;
	
	PROCEDURE g_Parameter(VAR x: g_Item; ftyp: g_Type; class: INTEGER; sc: BOOLEAN);
	  VAR r1, r2: LONGINT; s: String;
	BEGIN 
	IF (class = g_Par) AND (ftyp^.form IN [g_Array, g_Record]) THEN
	{ Nur die Basisadresse bergeben: }
	  IF x.i_type^.form = ftyp^.form THEN BEGIN
	    IF smallproc THEN BEGIN
	      r1 := 15 - ParCount;
	      IF x.lev = 0 THEN
	        writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', ','#(', hexword(ramstart + x.a - 2 - x.i_type^.size), ' + roffs)')
	      ELSE IF x.lev = curlev THEN
		IF sc THEN BEGIN
		  writeln(Taus, chr(9), 'mov',chr(9), 'R2, CP');
		  writeln(Taus, chr(9), 'sub', chr(9), 'R2, #', x.a);
		  writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', [R2]')
		END
		ELSE BEGIN 
		  { Da smallproc = TRUE ist patzt hier MakeItem. Es muss daher aus
		    der bergebenen Registernummer wieder der Offset zu FSP
		    berechnet werden: }
		  r2 := 2 * (16 - x.r) - 2 - x.i_type^.size;
		  writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', R', FSP);
		  writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', [R', r1, '+#', x.a - 2, ']');
		END
	      ELSE Mark('level!');
	    END
	    ELSE BEGIN { lange Prozedur TODO }    
	      r1 := ParCount * 2;
	      IF x.lev = 0 THEN BEGIN
		r2 := ramstart + x.a - 2- x.i_type^.size;
		writeln(Taus, chr(9), 'mov', chr(9), 'R2, ', '#)', hexword(r2), ' + roffs)')
	      END
	      ELSE IF x.lev = curlev THEN
		IF sc THEN BEGIN
		  writeln(Taus, chr(9), 'mov',chr(9), 'R2, CP');
		  writeln(Taus, chr(9), 'sub', chr(9), 'R2, #', 2 * x.a + 2);
		  writeln(Taus, chr(9), 'mov',chr(9), 'R2, [R2]');
		END
		ELSE BEGIN
		  writeln(Taus, chr(9), 'mov', chr(9), 'R2, R', FP);
		  IF x.a <> 0 THEN
		    writeln(Taus, chr(9), 'add', chr(9), 'R2, #', x.a);
		  writeln(Taus, chr(9), 'mov',chr(9), 'R2, [R2]');
		END
	      ELSE Mark('level!');
	      writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', r1, '], R2');
	    END
	  END
	  ELSE Mark('parametertype?')
	ELSE
	IF ((x.i_type^.form IN [g_Array, g_Record]) AND (x.i_type^.base^.form  = ftyp^.form))
	  OR (x.i_type = ftyp ) THEN BEGIN 
	  IF smallproc THEN BEGIN
	    r1 := 15 - ParCount;
	    IF x.mode = g_String THEN BEGIN
	      str(x.a, s);
	      writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', #String', s);
	    END
	    ELSE IF x.mode = g_Const THEN
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', #', hexword(x.a))
	    ELSE IF x.lev = 0 THEN BEGIN
	        r2 := ramstart + x.a - 2;  
	        write(Taus, chr(9), 'mov', chr(9), 'R', r1, ', ');
	        IF class = g_Par THEN 
	          writeln(Taus, '#(', hexword(r2), ' + roffs)')
	        ELSE IF class = g_Var THEN
	          writeln(Taus, '(', hexword(r2), ' + roffs)')
	        ELSE Mark('parametertype?');
	    END
	    ELSE IF sc THEN BEGIN
	      IF (x.mode = g_Var) OR (x.mode = g_Reg) Then Begin
	        writeln(Taus, chr(9), 'mov',chr(9), 'R2, CP');
	        writeln(Taus, chr(9), 'sub', chr(9), 'R2, #', x.a);
	        IF class = g_Par THEN
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', R2')
	        ELSE 
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', [R2]');
	      END
	      ELSE Mark('parametertype?');
	    END
	    ELSE BEGIN
	      { Da smallproc = TRUE ist patzt hier MakeItem. Es muss daher aus
	        der bergebenen Registernummer wieder der Offset zu FSP
	        berechnet werden: }
	      r2 := 2 * (16 - x.r) - 2;
	      IF (x.mode = g_Var) OR (x.mode = g_Reg) THEN BEGIN
	        IF class = g_Par THEN BEGIN
	          writeln(Taus, chr(9), 'mov', chr(9), 'R2, [R', FSP, '+#', r2, ']');
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', R2'); 
	        END 
	        ELSE
	          writeln(Taus, chr(9), 'mov', chr(9), 'R', r1, ', [R', FSP, '+#', r2, ']');
	      END
	      ELSE Mark('parametertype?');
	    END;
	  END
	  ELSE BEGIN { lange Prozedur }
	    r1 := ParCount * 2;
	    IF x.mode = g_String THEN BEGIN
	      str(x.a, s);
	      writeln(Taus, chr(9), 'mov', chr(9), 'R2, #String', s);
	      writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', r1, '], R2');
	    END
	    ELSE IF x.mode = g_Const THEN BEGIN
	      writeln(Taus, chr(9), 'mov', chr(9), 'R2, #', hexword(x.a));
	      writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', r1, '], R2');
	    END
	    ELSE IF x.lev = 0 THEN BEGIN
	      write(Taus, chr(9), 'mov', chr(9), 'R2, ');
	      r2 := ramstart + x.a - 2;
	      IF class = g_Par THEN
	        writeln(Taus, '#(', hexword(r2), ' +roffs)')
	      ELSE IF class = g_Var THEN
	        writeln(Taus, '(', hexword(r2), ' + roffs)')
	      ELSE Mark('parametertype?');
	      writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', r1, '], R2');
	    END
	    ELSE IF sc THEN BEGIN
	      IF (x.mode = g_Var) OR (x.mode = g_Reg) Then Begin
	        writeln(Taus, chr(9), 'mov',chr(9), 'R2, CP');
	        writeln(Taus, chr(9), 'sub', chr(9), 'R2, #', 2 * x.a);// + 2??? wg. MakeItem
	        IF class <> g_Par THEN
	          writeln(Taus, chr(9), 'mov', chr(9), 'R2, [R2]');
	        writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', r1, '], R2');
	      END
	      ELSE Mark('parametertype?');
	    END
	    ELSE BEGIN
	      IF (x.mode = g_Var) OR (x.mode = g_Reg) THEN BEGIN
	        writeln(Taus, chr(9), 'mov', chr(9), 'R2, R', FP);
	        IF x.a <> 0 THEN
		  writeln(Taus, chr(9), 'add', chr(9), 'R2, #', x.a);
	        IF class <> g_par THEN
	          writeln(Taus, chr(9), 'mov', chr(9), 'R2, [R2]');
	        writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', r1, '], R2');
	      END
	      ELSE Mark('parametertype?');
	    END;
	  END;
	END
	ELSE Mark('bad parameter type');
	END {g_Parameter};


	PROCEDURE g_OpenProc(blksize: INTEGER; small: BOOLEAN);
	  VAR i: INTEGER;
	BEGIN 
	  FOR i := 1 TO 15 DO BEGIN 
	    Regs[i] := -1; RegUsed[i] := 0; END;
	  writeln(Taus, chr(9), 'push', chr(9), 'R', FSP);
	  writeln(Taus, chr(9), 'add', chr(9), 'CP, #32');
	  writeln(Taus, chr(9), 'nop');
	  writeln(Taus, chr(9), 'pop', chr(9), 'R', FSP);
	  IF NOT small THEN BEGIN
	    writeln(Taus, chr(9), 'mov', chr(9), 'R', FP, ', R', FSP);
	    writeln(Taus, chr(9), 'sub', chr(9), 'R', FSP, ', #', blksize );
	  END;
	  If DebugLevel > 1 Then CheckStack	 
	END;
	
	PROCEDURE Call(VAR x: g_Object);
	  BEGIN writeln(Taus, chr(9), 'call', chr(9), modname + x^.name);
	  END {Call};
	
	PROCEDURE Return(s: string; int: BOOLEAN);
	VAR i: INTEGER;
	  BEGIN 
	    FOR i := 1 TO 15 DO BEGIN
	      IF RegUsed[i] <> 0 THEN Writeln('RegUsed[', i, '] = ', RegUsed[i]);
	      Regs[i] := -1; RegUsed[i] := 0; END;
	    writeln(Taus, chr(9), 'sub', chr(9), 'CP, #32');
	    IF int THEN
	      writeln(Taus, chr(9), 'reti', chr(9), chr(9), chr(9), chr(9), '; ', s)
	    ELSE
	      writeln(Taus, chr(9), 'ret', chr(9), chr(9), chr(9), chr(9), '; ', s);
	    //IF Sproc_Open THEN Mark('WriteLn?');
	  END {Return};
  
	PROCEDURE BJump(lab: LONGINT);
	  BEGIN write(Taus, chr(9), 'jmp', chr(9));
	        SetLabel(lab, '');
	  END {BJump};

	PROCEDURE Put_Read;
	  VAR lab: Longint; AscFlag: String;
	BEGIN Str(AscPort, AscFlag); 
	  AscFlag := 'S' + AscFlag + 'RIR';
	  lab := GetLabel; Newlabel(lab, '');
	  write(Taus, chr(9), 'jnb', chr(9), AscFlag, ', ');
	  setlabel(lab, '');
	  writeln(Taus, chr(9), 'bclr', chr(9), AscFlag);
	  //writeln(Taus, chr(9), 'mov', chr(9), 'R', nr, ', S0RBUF');
	END;


	PROCEDURE IOCall(VAR x, y: g_Item);
	  VAR nr: INTEGER;  adr: LONGINT; AscReg:String;
	  BEGIN Str(AscPort, AscReg); 
	    AscReg := 'S' + AscReg + 'RBUF';
	    IF x.a = 4 THEN BEGIN { WriteLn; }
	        Put_writeln; 
	    END
	    ELSE IF x.a = 1 THEN BEGIN { Read }
	      IF y.i_type^.form > 1 THEN
	        MARK('bad readparameter!');
	      Put_Read;
	      IF (y.mode = g_Var) OR (y.mode = g_Par) OR (y.mode = g_Reg) THEN BEGIN
	        IF y.lev = 0 THEN BEGIN
	          adr := Ramstart + y.a - 2;
	          IF DPP[3] <> 3 THEN 
		    writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
	          writeln(Taus, chr(9), 'mov', chr(9), hexword(adr), ', ', AscReg);
	        END
	        ELSE IF y.lev = curlev THEN BEGIN
	          IF smallproc THEN BEGIN
	            IF (y.mode = g_Var) OR (y.mode = g_Reg) THEN BEGIN
	              IF DPP[3] <> 3 THEN 
			writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
		      writeln(Taus, chr(9), 'mov', chr(9), 'R', y.r, ', ', AscReg); END
	            ELSE BEGIN
	              IF DPP[3] <> 3 THEN 
			writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
	              writeln(Taus, chr(9), 'mov', chr(9), '[R', y.r, '], ', AscReg)
	            END;
	          END
	          ELSE BEGIN
	            IF y.mode = g_Var THEN BEGIN
	              Getreg(nr); 
	              IF DPP[3] <> 3 THEN 
			writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
	              writeln(Taus, chr(9), 'mov', chr(9), 'R', nr, ', ', AscReg);
	              writeln(Taus, chr(9), 'mov', chr(9), '[R', FSP, '+#', y.a, '], R', nr)
	            END
	            ELSE BEGIN
	              Getreg(nr);
	              writeln(Taus, chr(9), 'mov', chr(9), 'R', nr, ', [R', FSP,'+#', y.a, ']');
	              IF DPP[3] <> 3 THEN 
			writeln(Taus, chr(9), 'extp', chr(9), '#3, #1');	
	              writeln(Taus, chr(9), 'mov', chr(9), '[R', nr, '], ', AscReg);
	            END;
	            y.r := nr; RetReg(y);
	          END;
	        END
	        ELSE Mark ('level?');
	      END
	      ELSE Mark('read where to?');
	    END
	    ELSE IF x.a = 5 THEN { Sleep }
	      writeln(Taus, chr(9), 'idle')
	    ELSE writeln('IOCall, ', y.tok);
	  END {IOCall};
	  
	  PROCEDURE OpenWrite;
	  BEGIN
	    write_used := TRUE;
	    Start_Sproc;
	    IF DebugLevel > 1 Then CheckStack;
	  END; {OpenWrite}
	  
	  PROCEDURE PutWrite(x: g_Item; form: Integer);
	  VAR adr, v: Longint;
	  BEGIN
	    IF x.mode = g_String THEN Write_String(x.a)
	    ELSE BEGIN
	        IF x.mode = g_Const THEN
	          writeln(Taus, chr(9), 'mov', chr(9), 'R1, #', x.a)
	        ELSE BEGIN
	          IF x.lev = 0 THEN BEGIN
	            adr := Ramstart + x.a - 2;
	            writeln(Taus, chr(9), 'mov', chr(9), 'R1, ', hexword(adr));
	          END 
	          ELSE BEGIN
	            IF (smallproc) THEN BEGIN 
	              writeln(Taus, chr(9), 'mov', chr(9), 'R1, CP');
	              v := 32 - 2 * x.r;
	              writeln(Taus, chr(9), 'sub', chr(9), 'R1, #', v);
	              writeln(Taus, chr(9), 'mov', chr(9), 'R1, [R1]');
	            END
	            ELSE BEGIN 
	              writeln(Taus, chr(9), 'mov', chr(9), 'R0, CP');
	              writeln(Taus, chr(9), 'sub', chr(9), 'R0, #32');
	              writeln(Taus, chr(9), 'mov', chr(9), 'R0, [R0]');
	              writeln(Taus, chr(9), 'add', chr(9), 'R0, #', x.a);
	              writeln(Taus, chr(9), 'mov', chr(9), 'R1, [R0]');
	            END;
	            IF x.mode = g_Par THEN
	              writeln(Taus, chr(9), 'mov', chr(9), 'R1, [R1]');
	          END;
	        END;
	          IF form = 0 THEN
	            writeln(Taus, chr(9), 'call', chr(9), 'writeval')
	          ELSE{ Hexadezimal }
	            writeln(Taus, chr(9), 'call', chr(9), 'writehex');
	      END;
	  END; {PutWrite}
	  
	  PROCEDURE CloseWrite(nl: Boolean);
	  BEGIN
	    IF nl THEN writeln(Taus, chr(9), 'call', chr(9), 'writeln');
	    end_Sprocs;  
	  END; {CloseWrite}	
	  
	PROCEDURE Init_Sproc(vsize:LONGINT);
	BEGIN
	  vartop := ramstart + vsize; ramtop := vartop;
	END;
	
	
  
	PROCEDURE Open;
	  BEGIN curlev := 0; pc := 0; //relx := 0; cno := 0;
	  END {Open};
	  
	  
	PROCEDURE mapram;
	VAR fpage, lpage,i: INTEGER;
	
	BEGIN fpage := ramstart DIV $4000; lpage := ramstop DIV $4000;	  
	  WHILE fpage <= lpage DO BEGIN
	    i := fpage MOD 4; DPP[i] := fpage;
	    fpage := fpage + 1;
	  END;
	  FOR i := 0 TO 3 DO 
	    writeln(Taus,chr(9), 'mov', chr(9), 'DPP', i, ', #', DPP[i]);
	  Write(Taus, chr(9), 'assume ');
	  FOR i := 0 TO 3 DO BEGIN
	    write(Taus, 'DPP', i, ':', DPP[i]);
	    IF i <> 3 THEN write(Taus, ', '); END;
	  Writeln(Taus);	
	END {mapram};
	
	PROCEDURE CallMain;
	BEGIN
	  writeln(Taus, 'roffs', chr(9), 'SET', chr(9), ramoffset);
	  IF Odd(DebugLevel)  Then BEGIN
	    writeln(Taus, ';@ #MainStart:', modname);
	    // Aktivieren, wenn das include monitor.asm eingefgt ist!
	    writeln(Taus, chr(9), 'jmp', chr(9), 'dbgmonitor'); 
	    END
	  ELSE
	    ;writeln(Taus, chr(9), 'jmp', chr(9), modname);
          //IF NOT impmod THEN writeln(Taus, '; End Of Init'); writeln(Taus);
	END;
	
	PROCEDURE Open_Gen;
	
	BEGIN
	  writeln(Taus, chr(9), 'cpu 80c167');
          writeln(Taus, chr(9), 'include reg166.inc');
          writeln(Taus, 'dpp3flag', chr(9), 'BIT', chr(9), 'PSW.6');
          
          IF progstart = 0 THEN BEGIN
            writeln(Taus,chr(9),'org',chr(9),'0200h',chr(9),chr(9),'; start at the end of vector-table');
            writeln(Taus, Vektoren[0].name, ':');
            IF NOT watchdog then
              writeln(Taus, chr(9),'diswdt',chr(9),chr(9), chr(9), '; watchdog off ');
            IF waitstates < 15 THEN BEGIN
              waitstates := 15 - waitstates;
              writeln(Taus, chr(9), 'or', chr(9), 'BUSCON0, #', waitstates, chr(9), '; set waitstates');
            END;
            writeln(Taus, chr(9), 'bset', chr(9), 'SYSCON.2', chr(9), '; enable XRAM');
            writeln(Taus, chr(9), 'bset', chr(9), 'SYSCON.8', chr(9), '; ClockOut on P3.15');
            writeln(Taus, chr(9), 'or', chr(9), 'SYSCON, #0E000h', chr(9), '; 1024 Bytes Stack');
            IF ConfFile ='' THEN
              writeln(Taus, chr(9), 'einit', chr(9), chr(9), chr(9), '; end of reset')
            ELSE
              writeln(Taus, chr(9), 'include', chr(9), ConfFile);
            { some more initialisations? Could be done via prog-parameters. }                        
            writeln(Taus, chr(9), 'mov', chr(9), 'STKUN, #', hexword(stackp));
            writeln(Taus, chr(9), 'mov', chr(9), 'STKOV, #', hexword(stackp - stacksize));
            writeln(Taus, chr(9), 'mov', chr(9), 'SP, #', hexword(stackp));
            writeln(Taus, chr(9), 'mov', chr(9), 'CP, #', hexword(contextp));
            writeln(Taus, chr(9), 'nop');
            writeln(Taus,chr(9), 'mov',chr(9), 'R', FSP, ', #',hexword(ramstop MOD $10000), chr(9),'; FSP = top of softstack');
            IF rammapped THEN mapram;
            IF Odd(DebugLevel) Then
	      writeln(Taus, chr(9), 'bset', chr(9), 'USR0', chr(9), chr(9), '; DebugFlag');
            writeln(Taus, '; End Of INIT'); writeln(Taus);
            
          END
          ELSE BEGIN
            writeln(Taus, chr(9), 'org', chr(9), hexword(progstart)); writeln(Taus);
            //writeln(Taus, chr(9), 'jmp', chr(9), modname); 
          END;
	END;
		
	PROCEDURE WriteVectors;
	  VAR i: INTEGER;
	BEGIN
	  writeln(Taus, chr(9), 'org', chr(9), '0000h');
	  writeln(Taus, chr(9), 'jmp', chr(9), Vektoren[0].name, chr(9), chr(9), '; Reset');
	  writeln(Taus, chr(9), 'org', chr(9), '0008h');
	  IF Vektoren[2].used THEN
	    writeln(Taus, chr(9), 'jmp', chr(9), Vektoren[2].name, chr(9), chr(9), '; NMI')
	  ELSE BEGIN
	    writeln(Taus, chr(9), 'bclr', chr(9), 'NMI');
	    writeln(Taus, chr(9), 'reti');
	  END;
	  writeln(Taus, chr(9), 'org', chr(9), '0010h');
	  IF Vektoren[4].used THEN
	    writeln(Taus, chr(9), 'jmp', chr(9), Vektoren[4].name, chr(9), chr(9), '; STKOF')
	  ELSE BEGIN
	    writeln(Taus, chr(9), 'bclr', chr(9), 'STKOF');
	    writeln(Taus, chr(9), 'reti');
	  END;
	  writeln(Taus, chr(9), 'org', chr(9), '0018h');
	  IF Vektoren[6].used THEN
	    writeln(Taus, chr(9), 'jmp', chr(9), Vektoren[6].name, chr(9), chr(9), '; STKUF')
	  ELSE BEGIN
	    writeln(Taus, chr(9), 'bclr', chr(9), 'STKUF');
	    writeln(Taus, chr(9), 'reti');
	  END;
	  writeln(Taus, chr(9), 'org', chr(9), '0028h');
	  IF Vektoren[10].used THEN
	    writeln(Taus, chr(9), 'jmp', chr(9), Vektoren[10].name, chr(9), chr(9), '; BTRAP')
	  ELSE BEGIN
	    writeln(Taus, chr(9), 'and', chr(9), 'TFR, #', hexword($FFFF - 143));
	    writeln(Taus, chr(9), 'reti');
	  END;
	  writeln(Taus, chr(9), 'org', chr(9), '0040h');
	  FOR i := 16 TO 71 DO BEGIN
	    IF VEKTOREN[i].used THEN BEGIN
	      writeln(Taus, chr(9), 'jmp', chr(9), vektoren[i].name);
	      writeln(Taus, chr(9), 'org', chr(9), hexword(4 * (i + 1)));
	    END
	    ELSE BEGIN
	      writeln(Taus, chr(9), 'reti', chr(9), chr(9), '; ', Vektoren[i].name);
	      writeln(Taus, chr(9), 'nop');
	    END
	  END;
	
	END;
	
	PROCEDURE End_MOD;
	  VAR lab: LONGINT;
	BEGIN
	  IF progstart = 0 THEN BEGIN
	    writeln(Taus); writeln(Taus, '; Endlosschleife:');
            lab := GetLabel; Newlabel(lab, '');
            write(Taus, chr(9), 'jmp', chr(9)); Setlabel(lab, '');
          END;
          Put_Strings;
          IF Write_Used THEN BEGIN
	    Put_Write;
	    IF (Odd(DebugLevel)) And (AscPort = 0)
	      THEN Mark(' No Writes while debugging!'); 
	  END;
                    
          IF Odd(DebugLevel) THEN BEGIN
	    writeln(Taus);
	    writeln(Taus, chr(9), 'include', chr(9), 'monitor.asm');
	    Vektoren[2].name := 'dbgmonitor'; Vektoren[2].used := TRUE; { NMI }
	    Vektoren[10].name := 'dbgmonitor'; Vektoren[10].used := TRUE; { BTRAP }
	    Vektoren[43].name := 'dbgmonitor'; Vektoren[43].used := TRUE; { S0RINT }
          END;
          If DebugLevel > 2 Then Begin
	    Vektoren[4].name := 'dbgmonitor'; Vektoren[4].used := TRUE; { STOTRAP }
          End;
          
          IF progstart = 0 THEN BEGIN
            writeln(Taus); writeln(Taus,'; Vektoren:');
            WriteVectors;
          END;
          Writeln('Ramtop: ', HexWord(ramtop));
	END;
	
	PROCEDURE InclMod(n: Ident);
	BEGIN
	  writeln(Taus,chr(9), 'include', chr(9), n + '.asm');
	END;

	PROCEDURE Enter_VEK(nr: INTEGER; name: STRING);
	BEGIN 
	  Vektoren[nr].name := name;
	  Vektoren[nr].used := FALSE;
	  IF nr <= 6 THEN Vektoren[nr].adr := nr * 4
	  ELSE IF nr = 10 THEN vektoren[nr].adr := 40
	  ELSE IF (nr > 15) AND (nr < 128) THEN
	    Vektoren[nr].adr := 64 + 4 * (nr - 16)
	  ELSE writeln('ungltiger Vektor');
	  //writeln(nr, ', ', name, ', ', Vektoren[nr].adr);
	END;

	

initialization

NEW(boolType); boolType^.form := g_Boolean; boolType^.size := 2;
NEW(intType); intType^.form := g_Integer; intType^.size := 2;
NEW(sfrType); sfrType^.form := g_SFRF; sfrType^.size := 2;
NEW(stringType); stringType^.form := g_stringf; stringType^.size := 2;
NextLab := 1; 
Nextreg := 1; 
Lastreg := 15;
maxreg := 0;
ramstart := $0E000; { Start of XRAM, might be overwritten }
ramstop := $0E7FE; { End of XRAM, might be overwritten }
progstart := 0; waitstates := 2;
stackp := $FE00; contextp := $F600; stacksize := 1024;
watchdog := FALSE; { watchdog disabled, might be changed in Sproc}
oplevel := 0;
smallproc := FALSE;
IsInt := FALSE; 
isproc := FALSE;
forloop := FALSE;
vartop := ramstart; { Top of global variables }
ramtop := vartop; { Top of used Ram abpve glob. vars in Sproc}
Lequ := NIL;
ConfFile := '';
DPP[0] := 0; DPP[1] := 1; DPP[2] := 2; DPP[3] := 3;
rammapped := FALSE; impmod := FALSE;
modname := ''; mainname := '';
ramoffset := 0;
write_used := FALSE;
DebugLevel := 0; AscPort := 0;

FOR vnr := 1 TO 15 DO BEGIN 
  Regs[vnr] := -1; RegUsed[i] := 0; END;

Enter_Vek(0,'RESET');
Enter_Vek(2, 'NMITRAP');
Enter_Vek(4, 'STOTRAP');
Enter_Vek(6, 'STUTRAP');
Enter_Vek(10, 'BTRAP');
FOR vnr := 16 TO 31 DO BEGIN
  str(vnr - 16, vstr); vstr := 'CC' + vstr + 'INT';
  Enter_Vek(vnr, vstr); END;
Enter_Vek(32, 'T0INT');
Enter_Vek(33, 'T1INT');
FOR vnr := 34 TO 38 DO BEGIN
  str(vnr - 32, vstr); vstr := 'T' + vstr + 'INT';
  Enter_Vek(vnr, vstr); END;
Enter_Vek(39, 'CRINT');
Enter_Vek(40, 'ADCINT');
Enter_Vek(41, 'ADEINT');
Enter_Vek(42, 'S0TINT');
Enter_Vek(43, 'S0RINT');
Enter_Vek(44, 'S0EINT');
Enter_Vek(45, 'SCTINT');
Enter_Vek(46, 'SCRINT');
Enter_Vek(47, 'SCEINT');
FOR vnr := 48 TO 60 DO BEGIN
  str(vnr - 32, vstr); vstr := 'CC' + vstr + 'INT';
  Enter_Vek(vnr, vstr); END;
Enter_Vek(61, 'T7INT');
Enter_Vek(62, 'T8INT');
Enter_Vek(63, 'PWMINT');
FOR vnr := 64 TO 67 DO BEGIN
  str(vnr - 64, vstr); vstr := 'XP' + vstr + 'INT';
  Enter_Vek(vnr, vstr); END;
FOR vnr := 68 TO 70 DO BEGIN
  str(vnr - 39, vstr); vstr := 'CC' + vstr + 'INT';
  Enter_Vek(vnr, vstr); END;
Enter_Vek(71, 'S0TBINT');


end.